/******************************************************************************
*  --- COVARIANCE ESTIMATOR ---                                               *
*                                                                             *
*  Copyright 2009 Bernhard Zeisl, Pierre Fite Georgel                         *
*                                                                             *
*  This program is free software: you can redistribute it and/or modify it    *
*  under the terms of the GNU General Public License as published by the Free *
*  Software Foundation, version 3 of the License.                             *
*  This program is distributed in the hope that it will be useful, but        *
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License    *
*  (LISENCE.txt) and the disclaimer (DISCLAIMER.txt) for more details.        *
*                                                                             *
*  The software was written by Bernhard Zeisl <bernhard.zeisl@mytum.de and    *
*  Pierre Fite Georgel <pierre.georgel@gmail.com>.                            *
*  For more information please contact the project website at:                *
*  http://campar.in.tum.de/Main/CovarianceEstimator                           *
*                                                                             *
*  The software package makes use of three additional libraries:              *
*  - The openCV computer vision library http://opencv.willowgarage.com/       *
*    published under the open source BSD license.                             *
*  - The SIFT implementation from Rob Hess <hess@eecs.oregonstate.edu>.       *
*    please see http://web.engr.oregonstate.edu/~hess/                        *
*    The package is distributed under GNU GPL v2 exept for the SIFT featur    *
*    detection code. For more information on SIFT licensing see LISENCE.ubc.  *
*  - The openSURF library from http://code.google.com/p/opensurf1/            *
*    The package is distributed under GNU GPL v3.                             *
*                                                                             *
*  If you use our software, please cite following publication:                *
*  Zeisl, B. et.al. "Estimation of Location Uncertainty for Scale Invariant   *
*    Feature Points", Proc. of the British Maschine Vision Conference (BMCV)  *
*    2009                                                                     *
*                                                                             *
*  @INPROCEEDINGS{ zeisl09elu,                                                *
*    AUTHOR={Bernhard Zeisl and Pierre Fite Georgel and Florian Schweiger and *
*            Eckehard Steinbach and Nassir Navab},                            *
*    TITEL={{Estimation of Location Uncertainty for Scale Invariant Feature   *
*            Points}},                                                        *
*    BOOKTITLE={Proc. of the British Maschine Vision Conference (BMCV)},      *
*    YEAR={2009}                                                              *
*  }                                                                          *
*                                                                             *
*******************************************************************************/

#include "CovEstimator.h"

using namespace std;

#define PI		3.14159265f


const IplImage* CovEstimator::getImageFromPyramid(int octv, int intvl) {

	/* Interpolation between layers based on the parameter subscale would be possible here,
	   but not implemented.
    */
	return detecPyr[octv][intvl];
}



CvMat* CovEstimator::getCovAt( float x, float y, float scale ) {

	// Retrieve the octave and interval the feature was detected at
	int octv=0, intv=0, row=0, col=0;
	float subintv=0.0f, subscale=0.0f;

	switch(this->type) {
		case DETECTOR_SIFT: {
			// scale calculation: scl = sigma * 2 ^ ((octv*intlvs + inv)/intvls) * 1/ 2;
			float tmp  = log(2 * scale / SIFT_SIGMA) / log(2.0f) * intervals;
			intv = ((int)cvRound(tmp) - 1) % intervals + 1;
			octv = ((int)cvRound(tmp) - 1) / intervals;
			subintv = tmp - (octv*intervals + intv);
			subscale = scale - SIFT_SIGMA/2 * pow(2.0f, (octv*intervals + intv)/intervals);

			// location calculation: feat->x = ( c + xc ) * pow( 2.0, octv );
			col = cvRound( x / pow(2.0, octv-1) );
			row = cvRound( y / pow(2.0, octv-1) );

			break;
		}
		case DETECTOR_SURF: {
			float size = scale * 9.0f/1.2f;
			if( size < 24 )
				octv = 0;
			else if( size < 48 ) // 45
				octv = 1;
			else if( size < 96 ) // 87
				octv = 2;
			else
				octv = 3;

			float tmp = (size/pow(2.0f, octv) - 9.0f) / 6.0f;
			intv = cvRound(tmp);
			subintv = tmp - intv;
			float subsize = static_cast<float>(size)
						- (9 + 6*static_cast<float>(intv))*pow(2.0f, octv);
			subscale = 1.2f * subsize / 9.0f; 

			col = cvRound( x / pow(2.0, octv) );
			row = cvRound( y / pow(2.0, octv) );

			break;
		}
			/* Information:
			   openCV in its current version (1.1 pre1a) does not implement the sizing of wavelets according to
			   the original paper from Bay et.al. openCV doubles the filter sizes for each octaves, whereas
			   in the original paper the sizing is a bit more complicated.
			   Assuming 4 octaves and 2 layers per octave the calculated filter sizes are as follows.
			   Underlined are the 2 layers which are used for 
			   
			   openCV:    9, 15, 21, 27
			   -------		 ------24
							  18,     30, 42, 54,
									  ------48
									   36,     60, 84, 108, 
											   ------96
												72,      120, 168, 216,
														 --------
			   calculations:
			   size = (9 + 6*intvl) * 2^oct
			   scale = 1.2 * (size) / 9.0

			   original paper from Bay et.al.:
			   ------------------------------
						  9, 15, 21, 27
							 ------24
							 15,     27, 39, 51
									 ------45
									 27,     51, 75, 99
											 ------87
											 51,     99, 147, 195
													 --------
			 
			   calculations:
			   ds = 6 * 2^oct
			   size = 9 + ( ds - 6 ) +  ( ds * intv ) = 9 - 6 + ds * (intv+1) = 3 + ds * (intv+1)
			   scale = 1.2 * (size + sub_size) / 9.0


			 --> Thus, the implementation of openCV needs to be changed, such that the original scales can
			     be used. If you want to do that, please see line 92 in ImagePyramid.cpp for changing the
				 covariances estimation accordingly.
				 Note, that then also the openCV libraries have to be updated (e.g. get the latest stable
				 snapshot from the SVN and compile openCV with the changes yourself)
			*/

			/* code for modification
			int sizes = { 9, 15, 21, 27,
							 15,     27, 39, 51,
								     27,     51, 75, 99,
											 51,     99, 147, 195 };
			 
			float ds = 6 * pow(2.0, octv);
			intv = cvRound( (size-3)/ds - 1 );
			subintv = ( (size-3)/ds - 1 ) - intv;
			int subsize = size - (3 + ds * (intv+1));
			subscale = 1.2 * subsize / 9.0;
			*/
	}

	// extract right image from pyramid
	const IplImage* img = getImageFromPyramid(octv, intv);

	// determine hessan at that point and calculate and scale covariance matrix
	this->H = hessian( img, row, col );
	cvInvert( this->H, this->cov, CV_SVD_SYM );

	// Hessian is estimated at particular octave and interval, thus scaling needed, which
	// adapts for the octave and subinterval
	/*** with norm: */
	// double norm = cvNorm( cov, 0, CV_L2 );
	// cvScale( cov, cov, pow(scale, 2) / norm );
	/*** or fixed value: */
	switch(this->type) {
		case DETECTOR_SIFT:
			cvScale( cov, cov, pow(2.0f, (octv + subintv/intervals) * 2) * 0.01 );
			break;
		case DETECTOR_SURF:
			cvScale( cov, cov, pow(1.2 * pow(2.0f, octv) + subscale, 2) * 10000);
			// filter size according to Bay et.al. paper:
			// float ds = 6 * pow(2.0, octv);
			// cvScale( cov, cov, pow(1.2/9 * (3 + ds) + subscale, 2) * 10000 );
			break;

	}
	
	cvSVD(this->cov, this->evals, this->evecs);
	ev1 = CV_MAT_ELEM(*evals, float, 0, 0);
	ev2 = CV_MAT_ELEM(*evals, float, 1, 0);
	if( ev1 < 0 && ev2 < 0 ) {
		ev1 = -ev1;
		ev2 = -ev2;
	}
	if( ev1 < ev2 ) {
		float tmp = ev1;
		ev1 = ev2;
		ev2 = tmp;
	}
	if( ev1 <= 0 || ev2 <= 0 ) {
		cout << "COV Eigenvalue of Hessian is negativ or zero(!)" << endl;
	}

	return this->cov;
}


void CovEstimator::drawCovInto(IplImage *img, int x, int y) {

	CvSize axes;
	axes = cvSize(static_cast<int>(ev1),
					static_cast<int>(ev2));
	float vx = CV_MAT_ELEM(*evecs, float, 0, 0);
	float vy = CV_MAT_ELEM(*evecs, float, 1, 0);
	float angle = sin( vy / sqrt(vx*vx + vy*vy) ) * 180/PI;
	cvEllipse( img,
			   cvPoint( x, y ), 
			   axes,
			   angle,
			   0,
			   360,
			   CV_RGB(0, 0, 0), 
			   2 );
	cvEllipse( img,
			   cvPoint( x, y ), 
			   axes,
			   angle,
			   0,
			   360,
			   CV_RGB(0, 255, 0), 
			   1 );
	// plot point
	// cvCircle( img, cvPoint(x,y), 1, CV_RGB(255, 0, 0) );
	cvLine( img, cvPoint(x-1, y-1), cvPoint(x+1, y+1), CV_RGB(0, 0, 0), 2 );
	cvLine( img, cvPoint(x-1, y+1), cvPoint(x+1, y-1), CV_RGB(0, 0, 0), 2 );
	cvLine( img, cvPoint(x-1, y-1), cvPoint(x+1, y+1), CV_RGB(0, 0, 255), 1 );
	cvLine( img, cvPoint(x-1, y+1), cvPoint(x+1, y-1), CV_RGB(0, 0, 255), 1 );
}


CvMat* CovEstimator::hessian( const IplImage* dog, int row, int col ) {

	int r, c;
/*	r = row; c = col;
	float v   =   pixval32f( dog, r, c );
	float dxx = ( pixval32f( dog, r, c+1 ) + 
				   pixval32f( dog, r, c-1 ) - 2 * v );
	float dyy = ( pixval32f( dog, r+1, c ) +
				   pixval32f( dog, r-1, c ) - 2 * v );
	float dxy = ( pixval32f( dog, r+1, c+1 ) -
				   pixval32f( dog, r+1, c-1 ) -
				   pixval32f( dog, r-1, c+1 ) +
				   pixval32f( dog, r-1, c-1 ) ) / 4.0;

*/	// ---
	float v, dxx = 0, dyy = 0, dxy = 0;
	float w[3][3] = { 0.0449f,    0.1221f,    0.0449f,
					  0.1221f,    0.3319f,    0.1221f,
					  0.0449f,    0.1221f,    0.0449f };

	for(int i = 0; i < 3; i++ )
		for( int j = 0; j < 3; j++ ) {
			r = row+(j-1); c = col+(i-1);
			v   =   pixval32f( dog, r, c );
			dxx += w[i][j] * ( pixval32f( dog, r, c+1 ) + 
							   pixval32f( dog, r, c-1 ) - 2 * v );
			dyy += w[i][j] * ( pixval32f( dog, r+1, c ) +
							   pixval32f( dog, r-1, c ) - 2 * v );
			dxy += w[i][j] * ( pixval32f( dog, r+1, c+1 ) -
							   pixval32f( dog, r+1, c-1 ) -
							   pixval32f( dog, r-1, c+1 ) +
							   pixval32f( dog, r-1, c-1 ) ) / 4.0f;

		}

	cvmSet( H, 0, 0, -dxx );
	cvmSet( H, 0, 1, -dxy );
	cvmSet( H, 1, 0, -dxy );
	cvmSet( H, 1, 1, -dyy );

	return H;
}