/******************************************************************************
*  --- COVARIANCE ESTIMATOR ---                                               *
*                                                                             *
*  Copyright 2009 Bernhard Zeisl, Pierre Fite Georgel                         *
*                                                                             *
*  This program is free software: you can redistribute it and/or modify it    *
*  under the terms of the GNU General Public License as published by the Free *
*  Software Foundation, version 3 of the License.                             *
*  This program is distributed in the hope that it will be useful, but        *
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License    *
*  (LISENCE.txt) and the disclaimer (DISCLAIMER.txt) for more details.        *
*                                                                             *
*  The software was written by Bernhard Zeisl <bernhard.zeisl@mytum.de and    *
*  Pierre Fite Georgel <pierre.georgel@gmail.com>.                            *
*  For more information please contact the project website at:                *
*  http://campar.in.tum.de/Main/CovarianceEstimator                           *
*                                                                             *
*  The software package makes use of three additional libraries:              *
*  - The openCV computer vision library http://opencv.willowgarage.com/       *
*    published under the open source BSD license.                             *
*  - The SIFT implementation from Rob Hess <hess@eecs.oregonstate.edu>.       *
*    please see http://web.engr.oregonstate.edu/~hess/                        *
*    The package is distributed under GNU GPL v2 exept for the SIFT featur    *
*    detection code. For more information on SIFT licensing see LISENCE.ubc.  *
*  - The openSURF library from http://code.google.com/p/opensurf1/            *
*    The package is distributed under GNU GPL v3.                             *
*                                                                             *
*  If you use our software, please cite following publication:                *
*  Zeisl, B. et.al. "Estimation of Location Uncertainty for Scale Invariant   *
*    Feature Points", Proc. of the British Maschine Vision Conference (BMCV)  *
*    2009                                                                     *
*                                                                             *
*  @INPROCEEDINGS{ zeisl09elu,                                                *
*    AUTHOR={Bernhard Zeisl and Pierre Fite Georgel and Florian Schweiger and *
*            Eckehard Steinbach and Nassir Navab},                            *
*    TITEL={{Estimation of Location Uncertainty for Scale Invariant Feature   *
*            Points}},                                                        *
*    BOOKTITLE={Proc. of the British Maschine Vision Conference (BMCV)},      *
*    YEAR={2009}                                                              *
*  }                                                                          *
*                                                                             *
*******************************************************************************/

#pragma once

//c++
#include <iostream>

//opencv
#include <cv.h>
#include <cxcore.h>

//covUtils
#include "definitions.h"

//sift
extern "C" {
	#include "utils.h"
	#include "sift.h"
}
class CovEstimator {

public:
	CovEstimator( const IplImage*** pyr, int type, int octvs, int intvls ) {
		this->type = type;
		octaves = octvs;
		intervals = intvls;
		detecPyr = pyr;
		H = cvCreateMat( 2, 2, CV_32FC1 );
		cov = cvCreateMat( 2, 2, CV_32FC1 );
		evals = cvCreateMat( 2, 1, CV_32FC1 );
		evecs = cvCreateMat( 2, 2, CV_32FC1 );
	};

	~CovEstimator() {
		cleanup();
	};

	const IplImage* getImageFromPyramid(int octv, int intvl);

	CvMat* getCovAt( float x, float y, float scale );

	void drawCovInto( IplImage* img, int x, int y );


private:
	/*** Methods ***/
	CvMat* hessian( const IplImage* img, int r, int c );

	/*  Linear interpolation
		target  - the target point, 0.0 - 1.0
		v       - a pointer to an array of size 2 containg the two values
	*/
	float linearInterp(float target, float v[]) {
		return (float)(target*(v[1])+ (1.0f - target)*(v[0]));
	}

	/*  BiLinear interpolation, linear interpolation in 2D
		target  - a 2D point (X,Y)
		v       - an array of size 4 containg values left to right, top to bottom
		cost: performs 3 linear interpolations
	*/
	float bilinearInterp(float target[], float v[])	{
		float v_prime[2] = { linearInterp(target[0], &(v[0])), 
							 linearInterp(target[0], &(v[2])) }; 
		return linearInterp(target[1], v_prime);
	}

	/*  TriLinear interpolation, linear interpolation in 3D
		target  - a 3D point (X,Y,Z)
		v       - an array of size 8 containg the values of the 8 corners 
				  of a cube defined as two faces: 0-3 face one (front face) 
                                                  4-7 face two (back face)
		cost: 7 linear interpolations
	*/
	float trilinearInterp(float target[], float v[]) {
		float v_prime[2] = { bilinearInterp(&(target[0]), &(v[0])),
							 bilinearInterp(&(target[0]), &(v[4])) };
		return linearInterp(target[2], v_prime);
	}

	void cleanup() {
		cvReleaseMat(&H);
		cvReleaseMat(&cov);
		cvReleaseMat(&evals);
		cvReleaseMat(&evecs);
	};


	/*** Member variables ***/
	int type;
	const IplImage*** detecPyr;
	int octaves, intervals;

	CvMat* H;
	CvMat* cov;
	CvMat* evals;
	float ev1, ev2;
	CvMat* evecs;
};
