/******************************************************************************
*  --- COVARIANCE ESTIMATOR ---                                               *
*                                                                             *
*  Copyright 2009 Bernhard Zeisl, Pierre Fite Georgel                         *
*                                                                             *
*  This program is free software: you can redistribute it and/or modify it    *
*  under the terms of the GNU General Public License as published by the Free *
*  Software Foundation, version 3 of the License.                             *
*  This program is distributed in the hope that it will be useful, but        *
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License    *
*  (LISENCE.txt) and the disclaimer (DISCLAIMER.txt) for more details.        *
*                                                                             *
*  The software was written by Bernhard Zeisl <bernhard.zeisl@mytum.de and    *
*  Pierre Fite Georgel <pierre.georgel@gmail.com>.                            *
*  For more information please contact the project website at:                *
*  http://campar.in.tum.de/Main/CovarianceEstimator                           *
*                                                                             *
*  The software package makes use of three additional libraries:              *
*  - The openCV computer vision library http://opencv.willowgarage.com/       *
*    published under the open source BSD license.                             *
*  - The SIFT implementation from Rob Hess <hess@eecs.oregonstate.edu>.       *
*    please see http://web.engr.oregonstate.edu/~hess/                        *
*    The package is distributed under GNU GPL v2 exept for the SIFT featur    *
*    detection code. For more information on SIFT licensing see LISENCE.ubc.  *
*  - The openSURF library from http://code.google.com/p/opensurf1/            *
*    The package is distributed under GNU GPL v3.                             *
*                                                                             *
*  If you use our software, please cite following publication:                *
*  Zeisl, B. et.al. "Estimation of Location Uncertainty for Scale Invariant   *
*    Feature Points", Proc. of the British Maschine Vision Conference (BMCV)  *
*    2009                                                                     *
*                                                                             *
*  @INPROCEEDINGS{ zeisl09elu,                                                *
*    AUTHOR={Bernhard Zeisl and Pierre Fite Georgel and Florian Schweiger and *
*            Eckehard Steinbach and Nassir Navab},                            *
*    TITEL={{Estimation of Location Uncertainty for Scale Invariant Feature   *
*            Points}},                                                        *
*    BOOKTITLE={Proc. of the British Maschine Vision Conference (BMCV)},      *
*    YEAR={2009}                                                              *
*  }                                                                          *
*                                                                             *
*******************************************************************************/
//C/C++
#include <iostream>
#include <string>

//openCV
#include <cv.h>
#include <cxcore.h>

//own
#include "ImagePyramid.h"
#include "CovEstimator.h"
#include "CovEstArgumentEvaluator.h"
#include "GeneralFeature.h"

// covUtils
#include "ipoint.h"
#include "covUtils.h"

// from SIFT detector
extern "C" {
	#include "sift.h"
	#include "imgfeatures.h"
}


int main(int argc, char* argv[])
{
	string filename;
	IplImage* frame = NULL, * drawFrame = NULL;

	/*** Parsing input arguments ***/
	CovEstArgumentEvaluator arg;
	arg.evaluate(argc, argv);


	/*** Loading input images and according keys ***/
	// Loading image
	filename.clear();
	filename.append(arg.imgDir).append(arg.imgFile);
	frame = cvLoadImage(filename.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
	if( !frame ){
		cout << "COV Error: Unable to load image from " << filename << "\n";
		exit( 1 );
	}

	// Loading the keypoints 
	IpVec ipts;
	filename.clear();
	filename.append(arg.imgDir).append(arg.keyFile);
	if(arg.verbose)
		cout << "COV Loading key points from file " << filename << endl;
	loadFeatures(filename.c_str(), ipts);
	

	/*** Creating image pyramid ***/
	int octaves=0, intervals=0;
	ImagePyramid pyr;
	IplImage*** imgPyr=NULL;

	switch (arg.decType) {
		case DETECTOR_SIFT: {
			// create image pyramid (modified functions from SIFT implementation)
			IplImage* init_img = create_init_img( frame, SIFT_IMG_DBL, SIFT_SIGMA );
			octaves = static_cast<int>(log(static_cast<float>(MIN( init_img->width, init_img->height )) ) / log(2.0f)) - 2;
			intervals = 3;
			IplImage*** gauss_pyr = build_gauss_pyr( init_img, octaves, intervals, SIFT_SIGMA );
			imgPyr = build_dog_pyr( gauss_pyr, octaves, intervals );
			release_pyr( &gauss_pyr, octaves, intervals + 3 );
			// file for storing resulting covariances
			filename.clear();
			filename.append(arg.imgDir).append(arg.imgFileStart).append(FE_GENERAL_KEYS_SIFT_COV);
			break;
		}
		case DETECTOR_SURF: {
			// create image pyramid
			octaves = 4; intervals = 2; // according to "current" (1.1 pre1a) openCV implementation
			pyr.create(frame, octaves, intervals);
			imgPyr = pyr.toArray();
			// file for storing resulting covariances
			filename.clear();
			filename.append(arg.imgDir).append(arg.imgFileStart).append(FE_GENERAL_KEYS_SURF_COV);
			break;
		}
	}
	ofstream& outfile = GeneralFeature::initializeFile(filename, ipts);


	/*** Estimating covariances for the keypoints ***/
	if(arg.verbose)
		cout << "COV Covariance estimation - Writing data to file " << filename << endl;

	// draw and write the data at the same time to reduce memory usage for matrix creation
	CovEstimator estimator( (const IplImage***) imgPyr, arg.decType, octaves, intervals );

	if(arg.saveCov || arg.showCov) {
		// Loading image for drawing
		filename.clear();
		filename.append(arg.imgDir).append(arg.imgFile);
		drawFrame = cvLoadImage(filename.c_str(), CV_LOAD_IMAGE_COLOR);
		if( !drawFrame ){
			cout << "COV Error: Unable to load image from " << filename << endl;
			exit( 1 );
		}

		// Do the covariance estimation
		for(unsigned int k = 0; k < ipts.size(); k++ ) {
			CvMat* cov = estimator.getCovAt( ipts.at(k).x, ipts.at(k).y, ipts.at(k).scale );
			estimator.drawCovInto( drawFrame, static_cast<int>(ipts.at(k).x),
												static_cast<int>(ipts.at(k).y));
			GeneralFeature::write(outfile, ipts.at(k), cov);
		}
	}
	else {
		// Do covariance estimation without drawing restults
		for(unsigned int k = 0; k < ipts.size(); k++ ) {
			CvMat* cov = estimator.getCovAt( ipts.at(k).x, ipts.at(k).y, ipts.at(k).scale );
			GeneralFeature::write(outfile, ipts.at(k), cov);
		}
	}
	GeneralFeature::closeFile(outfile);

	/*** Displaying image with covariances ***/
	if(arg.showCov) {
		showImage("Uncertainty Estimation: keypoints and their covariance", drawFrame);
		cvWaitKey(0);
	}


	/*** Save image with covariances to file ***/
	if(arg.saveCov) {
		filename.clear();
		filename.append(arg.imgDir).append(arg.imgFileStart);
		switch(arg.decType) {
			case DETECTOR_SIFT:
				filename.append(FE_COVARIANCE_IMAGE_SIFT);
				break;
			case DETECTOR_SURF:
				filename.append(FE_COVARIANCE_IMAGE_SURF);
				break;
		}
		cvSaveImage(filename.c_str(), drawFrame);
		if(arg.verbose)
			cout << "COV Image with covariances saved to file " << filename << endl;
	}


	/*** Clean up ***/
	cvReleaseImage(&frame);
	if(arg.saveCov || arg.showCov)
		cvReleaseImage(&drawFrame);
	
	if (arg.decType == DETECTOR_SIFT)
		release_pyr( &imgPyr, octaves, intervals + 2 );

	return 0;
}

