/******************************************************************************
*  --- COVARIANCE ESTIMATOR ---                                               *
*                                                                             *
*  Copyright 2009 Bernhard Zeisl, Pierre Fite Georgel                         *
*                                                                             *
*  This program is free software: you can redistribute it and/or modify it    *
*  under the terms of the GNU General Public License as published by the Free *
*  Software Foundation, version 3 of the License.                             *
*  This program is distributed in the hope that it will be useful, but        *
*  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
*  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License    *
*  (LISENCE.txt) and the disclaimer (DISCLAIMER.txt) for more details.        *
*                                                                             *
*  The software was written by Bernhard Zeisl <bernhard.zeisl@mytum.de and    *
*  Pierre Fite Georgel <pierre.georgel@gmail.com>.                            *
*  For more information please contact the project website at:                *
*  http://campar.in.tum.de/Main/CovarianceEstimator                           *
*                                                                             *
*  The software package makes use of three additional libraries:              *
*  - The openCV computer vision library http://opencv.willowgarage.com/       *
*    published under the open source BSD license.                             *
*  - The SIFT implementation from Rob Hess <hess@eecs.oregonstate.edu>.       *
*    please see http://web.engr.oregonstate.edu/~hess/                        *
*    The package is distributed under GNU GPL v2 exept for the SIFT featur    *
*    detection code. For more information on SIFT licensing see LISENCE.ubc.  *
*  - The openSURF library from http://code.google.com/p/opensurf1/            *
*    The package is distributed under GNU GPL v3.                             *
*                                                                             *
*  If you use our software, please cite following publication:                *
*  Zeisl, B. et.al. "Estimation of Location Uncertainty for Scale Invariant   *
*    Feature Points", Proc. of the British Maschine Vision Conference (BMCV)  *
*    2009                                                                     *
*                                                                             *
*  @INPROCEEDINGS{ zeisl09elu,                                                *
*    AUTHOR={Bernhard Zeisl and Pierre Fite Georgel and Florian Schweiger and *
*            Eckehard Steinbach and Nassir Navab},                            *
*    TITEL={{Estimation of Location Uncertainty for Scale Invariant Feature   *
*            Points}},                                                        *
*    BOOKTITLE={Proc. of the British Maschine Vision Conference (BMCV)},      *
*    YEAR={2009}                                                              *
*  }                                                                          *
*                                                                             *
*******************************************************************************/

// C++ headers
#include <iostream>
#include <string>
#include <algorithm>

// openCV headers
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

// own headers
#include "DecArgumentEvaluator.h"
#include "definitions.h"

// covUtils
#include "ipoint.h"
#include "covUtils.h"

// from SIFT detector
extern "C" {
#include "sift.h"
#include "imgfeatures.h"
}

using namespace std;

void cvSeq2Ipts(const CvSeq* keypoints, const CvSeq* descriptors, IpVec& ipts);
void feature2Ipts(struct feature* features, int nrFeat, IpVec& ipts);


int main(int argc, char* argv[]) {

	IplImage* grayImg, * colorImg;
	string filename;


	/*** Parsing input arguments ***/
	DecArgumentEvaluator arg;
	arg.evaluate(argc, argv);


	/*** Load the input image ***/
	filename.clear();
	filename.append(arg.imgDir).append(arg.imgFile);
	grayImg  = cvLoadImage(filename.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
	if(!grayImg) {
		cout << "DEC Error: Could not load image " << filename << "/n";
		exit(1);
	}

	
	/*** Detect feature points ***/
	if(arg.verbose)
		cout << "DEC Finding " << arg.decTypeName << " feature points..." << endl;

	IpVec ipts(arg.decType);

	switch( arg.decType ) {
		case DETECTOR_SIFT: {
			// using SIFT implementation from Rob Hess (http://web.engr.oregonstate.edu/~hess/)
			struct feature* features = NULL;
			int n = _sift_features(grayImg, &features, SIFT_INTVLS, SIFT_SIGMA, SIFT_THRES,
							SIFT_CURV_THR, SIFT_IMG_DBL, SIFT_DESCR_WIDTH, SIFT_DESCR_HIST_BINS);
			// convert found points to Ipoint structure
			feature2Ipts(features, n, ipts);
			delete [] features;
			break;
		}
		case DETECTOR_SURF: {
			// using SURF implementation from openCV
			CvSeq* keypoints   = NULL;
			CvSeq* descriptors = NULL;
			CvMemStorage* storage = cvCreateMemStorage(0);
			CvSURFParams params = cvSURFParams(SURF_THRES); 
			cvExtractSURF(grayImg, NULL, &keypoints, &descriptors, storage, params);
			// convert found points to Ipoint structure
			cvSeq2Ipts(keypoints, descriptors, ipts);
			cvClearMemStorage(storage);
			break;
		}
	}

	if(arg.verbose)
		cout << "DEC Found " << ipts.size() << " feature points." << endl;


	/*** Write back the found feature points to a file ***/
	// create default output for key file if not specicifed
	if(arg.keyFile.empty()) {
		arg.keyFile.append(arg.imgFileStart);
		switch( arg.decType ) {
			case DETECTOR_SIFT:
				arg.keyFile.append(FE_KEYS_SIFT);
				break;
			case DETECTOR_SURF:
				arg.keyFile.append(FE_KEYS_SURF);
				break;
		}
	}
	filename.clear();
	filename.append(arg.imgDir).append(arg.keyFile);
	saveFeatures(filename.c_str(), ipts);
	if(arg.verbose)
		cout << "DEC Feature points saved to file " << filename << endl;


	/*** Draw features in image, show and save if specified ***/
	if(arg.showPts || arg.savePts) {
		filename.clear();
		filename.append(arg.imgDir).append(arg.imgFile);
		colorImg = cvLoadImage(filename.c_str(), CV_LOAD_IMAGE_COLOR);
		drawIpoints(colorImg, ipts, arg.decType);
	
		if(arg.showPts) {
			string title(arg.decTypeName);
			title.append(": keypoints and their according scale");
			showImage(title.c_str(), colorImg);
		}
	

		if(arg.savePts) {
			filename.clear();
			filename.append(arg.imgDir).append(arg.imgFileStart);
			switch(arg.decType) {
				case DETECTOR_SIFT:
					filename.append(FE_FEATURES_IMAGE_SIFT);
					break;
				case DETECTOR_SURF:
					filename.append(FE_FEATURES_IMAGE_SURF);
					break;
			}
			cvSaveImage(filename.c_str(), colorImg);
			if(arg.verbose)
				cout << "DEC Image with feature points saved to file " << filename << endl;
		}
	}
	/*** Clean up ***/
	cvReleaseImage(&grayImg);
	if(arg.showPts || arg.savePts)
		cvReleaseImage(&colorImg);
	
	return 0;
}


/** Convert SURF feature points found by openCV to Ipoint structure
 */
void cvSeq2Ipts(const CvSeq* keypoints, const CvSeq* descriptors, IpVec& ipts) {

	/* from cv.h
	typedef struct CvSURFPoint
	{
		CvPoint2D32f pt;
		int laplacian;
		int size;
		float dir;
		float hessian;
	} CvSURFPoint;
	*/

	// empty vector of interest points
	ipts.clear();

	CvSURFPoint* surfPt;
	float* descr;
	Ipoint* ipt;
	for( int k = 0; k < keypoints->total; k++ ) {
		
		// get point from sequence
		surfPt = (CvSURFPoint*) cvGetSeqElem(keypoints, k);

		// discard if scale value is impossible
		if(surfPt->size < 9)
			continue;
		
		// create new interest point and assign values
		ipt = new Ipoint();
		ipt->x			 = surfPt->pt.x;
		ipt->y			 = surfPt->pt.y;
		ipt->laplacian	 = surfPt->laplacian;
		ipt->orientation = surfPt->dir;
		ipt->scale		 = (float) (1.2/9.0 * surfPt->size);

		// assign descriptor data
		descr = (float*) cvGetSeqElem(descriptors, k);
		ipt->initializeDescr(SURF_DESCR_SIZE);
		for( int i = 0; i < SURF_DESCR_SIZE; i++ )
			ipt->descriptor[i] = descr[i];

		// add to vector of interest points
		ipts.push_back(*ipt);
	}
}

/** Convert SIFT feature points found by Rob Hess SIFT implementation to Ipoint structure
 */
void feature2Ipts(struct feature* features, int nrFeat, IpVec& ipts) {
	ipts.clear();
	Ipoint* ipt;
	for( int k = 0; k < nrFeat; k++ ) {
		// create new interest point and assign values
		ipt = new Ipoint();
		ipt->x			 = (float) features[k].x;
		ipt->y			 = (float) features[k].y;
		ipt->laplacian	 = 0;
		ipt->orientation = (float) features[k].ori;
		ipt->scale		 = (float) features[k].scl;

		// assign descriptor data
		ipt->initializeDescr(features[k].d);
		for( int i = 0; i < features[k].d; i++ )
			ipt->descriptor[i] = (float) features[k].descr[i];

		// add to vector of interest points
		ipts.push_back(*ipt);
	}
}